//-----------------------------------------------------------------------------
// ga.cpp
//-----------------------------------------------------------------------------
// 
//  GA for grouping problem 
//
//-----------------------------------------------------------------------------

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
 

#include <stdexcept>  
#include <iostream>   
#include <fstream>
#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>  
#endif



#include <eo>
#include <ga.h>

#include "operators.h"
#include "fitness.h"

// #include "gui/gui.h"



using namespace std;





void main_function(int argc, char **argv)
{


// PARAMETRES
   unsigned int SEED;      // seed for random number generator
   unsigned int POP_SIZE;  // Size of population
   unsigned int MAX_GEN;  // Maximum number of generation before STOP
   double CROSS_RATE;      // Crossover rate
   double P_MUT_PER_BIT; // probability of bit-flip mutation
   double MUT_RATE;        // mutation rate
   double P_MUT1, P_MUT2;
   int K; // # hromadek 

// PARAMETRES BY USER 

     eoParser parser(argc, argv);

 
    eoValueParam<uint32_t> seedParam(time(0), "seed", "Random number seed", 'S');
    parser.processParam( seedParam );
    SEED = seedParam.value();

   
   // parameters for evolution engine
    eoValueParam<unsigned int> popSizeParam(50, "popSize", "Population size",'P');
    parser.processParam( popSizeParam, "Evolution engine" );
    POP_SIZE = popSizeParam.value();

    // init and stop
    eoValueParam<unsigned int> maxGenParam(10000, "maxGen", "Maximum number of generations",'G');
    parser.processParam( maxGenParam, "Stopping criterion" );
    MAX_GEN = maxGenParam.value();

   // operators probabilities at the algorithm level
    eoValueParam<double> pCrossParam(0.5, "pCross", "Probability of Crossover", 'C'); 
    parser.processParam( pCrossParam, "Genetic Operators" );
    CROSS_RATE = pCrossParam.value();

    eoValueParam<double> pMutParam(0.5, "pMut", "Probability of Mutation", 'M');
    parser.processParam( pMutParam, "Genetic Operators" );
    MUT_RATE = pMutParam.value();

   // relative rates and private parameters for mutations;
    eoValueParam<double> pMutPerBitParam(0.01, "pMutPerBit", "Probability of flipping 1 bit in bit-flip mutation", 'b');
    parser.processParam( pMutPerBitParam, "Genetic Operators" );
    P_MUT_PER_BIT = pMutPerBitParam.value();


    eoValueParam<double> pMut1(1.0, "pMutace1", "Pst stare mutace", 'b');
    parser.processParam( pMut1, "Mutace" );
    P_MUT1 = pMut1.value();

    eoValueParam<double> pMut2(0.0, "pMutace2", "Pst nove mutace", 'b');
    parser.processParam( pMut2, "Druha mutace" );
    P_MUT2 = pMut2.value();

    
    // name of file to save statistics 
    eoValueParam<string> saveStatsParam("fitness.log", "LogFile","A file to log statistics",'L');
    parser.processParam( saveStatsParam, "Log Statistics File" );
    string saveStatsName = saveStatsParam.value();


    // the name of the "status" file where all actual parameter values will be saved
    string str_status = parser.ProgramName() + ".status"; // default value
    eoValueParam<string> statusParam(str_status.c_str(), "status","Status file",'S');
    parser.processParam( statusParam, "Persistence" );

    
    // zadani
    eoValueParam<string>  inputParam("zadani.txt", "InputFile","A file with input",'I');
    parser.processParam( inputParam, "Input File" );
    string inputName = inputParam.value();
    
    eoValueParam<int> KParam(10,"NumberOfGroups","Number of Groups",'K');
    parser.processParam( KParam, "NumberOfGroups");
    K = KParam.value(); 
    
    if (parser.userNeedsHelp())
      {
        parser.printHelp(cout);
        exit(1);
      }
    if (statusParam.value() != "")
      {
	ofstream os(statusParam.value().c_str());
	os << parser;		// and you can use that file as parameter file
      }
    

    /// read input file
    ifstream fin(inputName.c_str());
    unsigned N;
    fin >> N; 

    
    //////////////////////////
    //  Random seed
    //////////////////////////
    rng.reseed(SEED);

    /////////////////////////////
    // Fitness function
    ////////////////////////////
    MyFitness eval(inputName,K);


  ////////////////////////////////
  // Initilisation of population
  ////////////////////////////////
  eoPop<MyIndividual> pop;
  for (unsigned int igeno=0; igeno<POP_SIZE; igeno++)
    {
      MyIndividual v;           // void individual, to be filled
      v.resize(N);
      for (unsigned ivar=0; ivar<N; ivar++)
        {
          int r = rng.random(K);  // new value, random in {0,K}
          v[ivar] = r;      
        }
      eval(v);                 // evaluate it !!!
      pop.push_back(v);        
    }

  // OUTPUT the initial population 
  //  pop.sort();
  // cout << "Initial Population" << endl;
  // cout << pop;
  cout << "START" << endl; 
   
 
  /////////////////////////////////////
  // selection and replacement
  ////////////////////////////////////
   eoDetTournamentSelect<MyIndividual> selectOne;  // turnajova selekce 
  // eoProportionalSelect<MyIndividual> selectOne;   // ruletova selekce 
  eoSelectPerc<MyIndividual> select(selectOne); 
  

  // REPLACE
  double elita_percentage = 0.05; 
  eoDeterministicSaDReplacement<MyIndividual> replace(elita_percentage,
						      1.0-elita_percentage);
  
  // CROSSOVER
  MyXover xover;

  // MUTATION
  MyMut mutation(K); 

  /// operatory je mozne kombinovat 
  /// napr. pouzit dva druhy mutaci  
  /// 
  //    MyMut2 mutation2(inputName,K);
  //   MyMut mutation1(K);
  // 
  /// kombinovana mutace - s psti P_MUT1 se pouzije mutation1, 
  ///                      s psti P_MUT2 se pouzije mutation2 
  //           ( P_MUT1 + P_MUT2 = 1.0; jinak jsou psti proporcionalne) 
  // 
  // eoPropCombinedMonOp<MyIndividual> mutation(mutation1,P_MUT1);
  // mutation.add(mutation2,P_MUT2);

  // Genetic operators are  encapsulated into an eoTRansform object
  eoDynSGATransform<MyIndividual> transform(xover, &CROSS_RATE, 
					    mutation, &MUT_RATE);


  // CHECKPOINT

  // stop after MAX_GEN generations
  eoGenContinue<MyIndividual> genCont(MAX_GEN);
  // eoFitContinue<MyIndividual> fitCont(0);
  eoCheckPoint<MyIndividual> checkpoint( genCont );    

  //myGui<MyIndividual> gui(&CROSS_RATE,&MUT_RATE,-300,0); // stopping via gui 

  //  eoCheckPoint<MyIndividual> checkpoint(gui);    

  // Create a counter parameter
  eoValueParam<unsigned> generationCounter(0, "Gen.");
  eoIncrementor<unsigned> increment(generationCounter.value());
  checkpoint.add(increment);
  
  eoBestFitnessStat<MyIndividual> bestStat;
  eoSecondMomentStats<MyIndividual> SecondStat;
  checkpoint.add(bestStat);
  checkpoint.add(SecondStat);
    

  eoFileMonitor fileMonitor(saveStatsName, " ");
  checkpoint.add(fileMonitor);
  fileMonitor.add(generationCounter);
  fileMonitor.add(bestStat);
  fileMonitor.add(SecondStat);

  
  /////////////////////////////////////////
  // the GA algorithm
  ////////////////////////////////////////
  eoEasyEA<MyIndividual> gga(checkpoint, 
		     eval, 
		     select, 
		     transform, 
		     replace);
  gga(pop);
  
  // OUTPUT the final population 
  //  pop.sort();
  // cout << "FINAL Population\n" << pop[0] << endl;
  cout << "END" << endl; 

  // uloz reseni do souboru 
  string outputName = inputName+".sol"; 
  ofstream out(outputName.c_str()); 
  out << N << endl; 
  double w; 
  for (unsigned i=0; i<N; i++) {
    fin >> w; 
    out << w << " " << pop.best_element()[i] << endl; 
  } 
}


int main(int argc, char **argv)
{

    try
    {
        main_function(argc, argv);
    }
    catch(exception& e)
    {
        cout << "Exception: " << e.what() << '\n';
    }

    return 1;
}
